@@ -27,20 +27,24 @@ pqxx::CachedResult::CachedResult(pqxx::TransactionItf &Trans,
   m_Cursor(Trans, Query, BaseName, Granularity),
   m_Pos(0), 
   m_Size(-1),
+  m_Lower(-1),
   m_Upper(-1)
 {
   // We can't accept granularity of 1 here, because some block number 
   // arithmetic might overflow.
   if (m_Granularity <= 1)
     throw out_of_range("Invalid CachedResult granularity");
+
+  if (m_Lower > 0)
+    throw logic_error("Internal libpqxx error: "
+	              "CachedResult::size_type is unsigned");
+
+  m_Upper = BlockFor(numeric_limits<size_type>::max()) + 1;
 }
 
 
 void pqxx::CachedResult::clear()
 {
-  m_Pos = -1;
-  m_Size = -1;
-  m_Upper = -1;
   m_Cache.clear();
 }
 
@@ -75,20 +79,28 @@ void pqxx::CachedResult::MoveTo(blocknum Block) const
 pqxx::Result pqxx::CachedResult::Fetch() const
 {
   Result R;
-  if ((m_Upper != -1) && (m_Pos >= m_Upper)) return R;
+  if (m_Pos == -1) 
+    throw logic_error("Internal libpqxx error: "
+	              "CachedResult fetches data from unknown cursor position");
+
+  if (m_Pos >= m_Upper) return R;
 
   try
   {
     if (!(m_Cursor >> R)) 
     {
       m_Upper = m_Pos;
-      if ((m_Size == -1) && (m_Pos == m_Cache.rend()->first + 1)) 
+      if ((m_Size == -1) && (m_Pos == (m_Cache.rend()->first + 1))) 
+      {
+        m_Lower = m_Pos - 1;
         m_Size = m_Pos * m_Granularity;
+      }
 
       m_Pos = -1;
       return R;
     }
     m_Cache.insert(make_pair(m_Pos, CacheEntry(R)));
+    if (m_Pos > m_Lower) m_Lower = m_Pos;
     if ((m_Size == -1) && (R.size() < m_Granularity))
     {
       // This is the last block.  While we're here, record result set size.
@@ -128,46 +140,19 @@ void pqxx::CachedResult::DetermineSize() const
     }
   }
 
-  if (m_Upper == -1)
+  const blocknum TopBlock = BlockFor(numeric_limits<size_type>::max());
+  if (m_Upper > TopBlock)
   {
-    const blocknum TopBlock = BlockFor(Cursor::ALL());
     Result T = GetBlock(TopBlock);
     if (!T.empty())
     {
-      m_Size = TopBlock*m_Granularity + T.size();
       if (m_Size <= 0)
 	throw runtime_error("Result set too large");
-
-      m_Upper = TopBlock + 1;
       return;
     }
   }
 
-  blocknum Lower = m_Cache.rbegin()->first;
-
-  while (Lower < m_Upper)
-  {
-    blocknum Middle = (Lower + m_Upper) / 2;
-
-    const size_type BlockSize = GetBlock(Middle).size();
-
-    if (BlockSize < m_Granularity)
-    {
-      // Found new upper bound to result set size
-      m_Upper = Middle;
-      if (BlockSize)
-      {
-        // Got a partial block--final row must be right here
-        ++m_Upper;
-        Lower = m_Upper;	// Terminate
-      }
-      // (Otherwise: Got empty block, probably overshot end-of-data)
-    }
-    else 
-    {
-      Lower = Middle + 1;
-    }
-  }
+  while ((m_Lower+1) < m_Upper) GetBlock((m_Lower + m_Upper)/2);
 
   CacheMap::reverse_iterator L = m_Cache.rbegin();
   m_Size = L->first * m_Granularity + L->second.Data().size();
